package control;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.HashMap;
import java.util.Observable;

import javax.swing.JFileChooser;

import view.ParameterPanel;

import model.*;
import model.fillingFunction.*;
import model.innerSolver.SolomonHeuristic;

/**
 * MVC Control component for the simulation.
 * 
 * The Control class implements the ActionListener interface. This makes this class a candidate for
 * use as ActionListener in swing components.
 * 
 * @author Thomas Kjeldsen
 */

public class InteractiveController extends Observable implements ActionListener
{

	private static final HashMap<String, Integer> commandMap = new HashMap<String, Integer>();
	private TabuSearch ts;
	private ParameterPanel parameterPanel;

	private JFileChooser fc;

	static
	{
		int commandNumber = 0;
		String[] commands = { "start", "step", "stop", "new", "load", "quit" };
		for (String command : commands)
			commandMap.put(command, commandNumber++);
	}

	/**
	 * Set up the needed parts of the controller.
	 * 
	 * @param ts
	 *          TabuSearch algorithm to observe.
	 */
	public InteractiveController (TabuSearch ts, ParameterPanel parameterPanel)
	{
		this.ts = ts;
		this.parameterPanel = parameterPanel;
		fc = new JFileChooser();
	}

	/**
	 * The method to be invoked by the swing components to handle event.
	 * 
	 * @param ev
	 *          The event received
	 */
	public void actionPerformed (ActionEvent ev)
	{
		Integer index = commandMap.get(ev.getActionCommand());
		Thread t = new Thread(ts);
		t.start();
		switch (index != null ? index : -1)
		{
		case 0: // start
			ts.setIsFullRun(true);
			break;
		case 1: // step
			ts.setIsStepping(true);
			break;
		case 2: // stop
			ts.setIsFullRun(false);
			ts.setIsStepping(false);
			break;
		case 3: // new
			if (fc.getSelectedFile() == null)
				fc.showOpenDialog(fc.getParent());

			constructNewTabuSearch(fc.getSelectedFile(), parameterPanel.getK_max());
			break;
		case 4: // load
			fc.showOpenDialog(fc.getParent());
			constructNewTabuSearch(fc.getSelectedFile(), parameterPanel.getK_max());
			break;
		case 5: // quit
			System.exit(0);
			break;
		default:
			break;
		}
	}

	private void constructNewTabuSearch (File file, int k_max)
	{
		DataReader dr = new DataReader(file);
		Customers customers = new Customers(dr);

		AbstractFillingFunction testFillingFunction = new InitialFillingFunction(customers, 1.0);
		SolomonHeuristic sh = new SolomonHeuristic(customers, dr.getCapacity(), testFillingFunction);
		Solution solution = sh.GetInitialSolution();

		ts = new TabuSearch(dr, solution, sh, testFillingFunction, 7, 3, k_max, 2, false);
		sendNotification();
	}

	protected void sendNotification ()
	{
		setChanged();
		notifyObservers();
	}

	/**
	 * @return the ts
	 */
	public TabuSearch getTs ()
	{
		return ts;
	}

	/**
	 * @param parameterPanel
	 *          the parameterPanel to set
	 */
	public void setParameterPanel (ParameterPanel parameterPanel)
	{
		this.parameterPanel = parameterPanel;
	}
}
