package model;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class Customers
{
	private int[][] customers;
	private List<Integer> customerList;
	private boolean customerListCache = false;
	private double[][] distances;
	private boolean isCalculated = false;
	private int numberOfCustomers;
	private int totalDemand;
	private boolean totalDemandCache = false;

	/**
	 * Class to contain and query information about customers. Constructor getting information about
	 * customers from DataReader class.
	 * 
	 * @param reader
	 *          The DataReader used to obtain information about data.
	 * @throws IOException
	 *           Throws an IOException if data is not read probably.
	 */
	public Customers (DataReader reader)
	{
		numberOfCustomers = reader.getNumberOfCustomers();
		customers = reader.getData();
	}

	/**
	 * @return the numberOfCustomers
	 */
	public int getNumberOfCustomers ()
	{
		return numberOfCustomers;
	}

	/**
	 * Returns a list of the customers in the Customers class.
	 * 
	 * @return List of the customers in the Customers class.
	 */
	public List<Integer> getListOfCustomers ()
	{
		if (!customerListCache)
		{
			customerList = new ArrayList<Integer>(numberOfCustomers);
			for (int i = 1; i < numberOfCustomers; i++)
				customerList.add(i);
		}

		return customerList;
	}

	/**
	 * Returns the distance between customer 1 and customer 2.
	 * 
	 * @param cust1
	 *          Customer 1.
	 * @param cust2
	 *          Customer 2.
	 * @return
	 */
	public double Distance (int cust1, int cust2)
	{
		if (!isCalculated)
			CalculateDistances();

		return distances[cust1][cust2];
	}

	/**
	 * Returns the distance between the depot and a customer.
	 * 
	 * @param customer
	 *          Customer 1.
	 * @return
	 */
	public double Distance (int customer)
	{
		if (!isCalculated)
			CalculateDistances();

		return distances[0][customer];
	}

	private void CalculateDistances ()
	{
		distances = new double[numberOfCustomers][numberOfCustomers];

		for (int i = 0; i < numberOfCustomers; i++)
		{
			for (int j = 0; j < numberOfCustomers; j++)
			{
				distances[i][j] = CalculateDistance(i, j);
			}
		}

		isCalculated = true;
	}

	/**
	 * Returns the cartesian distance between two customers.
	 * 
	 * @param i
	 *          Customer i.
	 * @param j
	 *          Customer j.
	 * @return Cartisian distance.
	 */
	private double CalculateDistance (int i, int j)
	{
		return Math.sqrt(Math.pow(Math.abs(customers[0][i] - customers[0][j]), 2)
				+ Math.pow(Math.abs(customers[1][i] - customers[1][j]), 2));
	}

	/**
	 * Gets the coordinates of a customer (or the depot).
	 * 
	 * @param customer
	 *          The customer whose coordinates is desired.
	 * @return Return int[2] with x and y coordinate.
	 */
	public int[] getCustomerCoordinates (int customer)
	{
		int[] coordinates = new int[2];
		coordinates[0] = customers[0][customer];
		coordinates[1] = customers[1][customer];

		return coordinates;
	}

	/**
	 * Returns demand of customer.
	 */
	public int getDemand (int customer)
	{
		return customers[2][customer];
	}

	/**
	 * Returns ready time of customer.
	 */
	public int getReadyTime (int customer)
	{
		return customers[3][customer];
	}

	/**
	 * Returns due date of customer.
	 */
	public int getDueDate (int customer)
	{
		return customers[4][customer];
	}

	/**
	 * Returns service time of customer.
	 */
	public int getServiceTime (int customer)
	{
		return customers[5][customer];
	}

	/**
	 * Clear the cache, that is resets distances table.
	 */
	public void clearCache ()
	{
		distances = new double[numberOfCustomers][numberOfCustomers];
		isCalculated = false;
		customerListCache = false;
		totalDemandCache = false;
	}

	/**
	 * Returns the total demand of all customers.
	 * 
	 * @return Total demand.
	 */
	public int getTotalDemand ()
	{
		if (!totalDemandCache)
		{
			totalDemand = 0;
			for (int j = 0; j < numberOfCustomers; j++)
				totalDemand += getDemand(j);
		}

		return totalDemand;
	}
}
