package model;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.StreamTokenizer;

/**
 * A class that can read data in the Solomon problem format.
 * 
 * @author jerome
 * 
 */
public class DataReader
{
	private String filePath;
	private int[][] data;
	private int numberOfCustomers;
	private int numberOfVehicles;
	private int capacity;
	private boolean didRead = false;

	/**
	 * Initializes the DataReader with a given path.
	 * 
	 * @param path
	 *          File path to read from.
	 */
	public DataReader (String path)
	{
		this.filePath = path;
		this.numberOfCustomers = queryNumberOfCustomers();
		data = new int[6][numberOfCustomers];
		readData();
	}

	/**
	 * Initializes the DataReader with a given file.
	 * 
	 * @param file
	 *          File to read.
	 */
	public DataReader (File file)
	{
		if (!file.canRead())
			throw new IllegalArgumentException("Cannot read data file: " + file.getAbsolutePath());

		this.filePath = file.getAbsolutePath();
		this.numberOfCustomers = queryNumberOfCustomers();
		data = new int[6][numberOfCustomers];
		readData();
	}

	/**
	 * Reads VRPTW data from the specified file.
	 * 
	 * @throws IOException
	 *           If data does not fit the predefined pattern an exception is thrown.
	 */
	private void readData ()
	{
		try
		{
			LineNumberReader lnr = new LineNumberReader(new FileReader(filePath));
			StreamTokenizer stok = new StreamTokenizer(lnr);
			stok.parseNumbers();
			stok.eolIsSignificant(true);
			stok.nextToken();

			// First two numbers of file are vehicle number and capacity.
			while (stok.ttype != StreamTokenizer.TT_NUMBER)
				stok.nextToken();

			numberOfVehicles = (int) stok.nval;
			stok.nextToken();
			capacity = (int) stok.nval;

			int line = 0;

			// Customer data are the next numbers.
			stok.nextToken();
			while (stok.ttype != StreamTokenizer.TT_NUMBER)
				stok.nextToken();

			while (stok.ttype != StreamTokenizer.TT_EOF)
			{
				stok.nextToken();// Skips the first entry as it is the customer number.
				int index = 0;

				while (stok.ttype != StreamTokenizer.TT_EOL)
				{
					if (stok.ttype == StreamTokenizer.TT_NUMBER)
						data[index][line] = (int) stok.nval;
					else
						throw new IOException("Corrupt input data file! Contains non-numbers or empty lines at end of file.");

					index++;
					stok.nextToken();
				}

				line++;
				stok.nextToken();
			}

			if (line != numberOfCustomers)
				throw new IOException("File does not contain the expected number of customers!");
		} catch (IOException e)
		{
			System.out.println(e.toString());
		}
		didRead = true;
	}

	/**
	 * Reads the file to determine the number of customers (including the depot).
	 * 
	 * @return Number of customers contained in file.
	 */
	private int queryNumberOfCustomers ()
	{
		int line = 0;
		try
		{
			LineNumberReader lnr = new LineNumberReader(new FileReader(filePath));
			StreamTokenizer stok = new StreamTokenizer(lnr);
			stok.parseNumbers();
			stok.eolIsSignificant(true);
			stok.nextToken();

			// First two numbers of file are vehicle number and capacity.
			while (stok.ttype != StreamTokenizer.TT_NUMBER)
				stok.nextToken();

			stok.nextToken();

			// Customer data are the next numbers.
			stok.nextToken();
			while (stok.ttype != StreamTokenizer.TT_NUMBER)
				stok.nextToken();

			while (stok.ttype != StreamTokenizer.TT_EOF)
			{
				while (stok.ttype != StreamTokenizer.TT_EOL)
				{
					if (stok.ttype != StreamTokenizer.TT_NUMBER)
						throw new IOException("Corrupt input data file! Contains non-numbers.");

					stok.nextToken();
				}

				line++;
				stok.nextToken();
			}

		} catch (IOException e)
		{
			System.out.println(e.toString());
		}

		return line;
	}

	/**
	 * Data format: (x, y, demand, readyTime, dueTime, serviceTime)
	 * 
	 * @return the data
	 */
	public int[][] getData ()
	{
		return data;
	}

	/**
	 * @return the numberOfVehicles
	 */
	public int getNumberOfVehicles ()
	{
		return numberOfVehicles;
	}

	/**
	 * @return the capacity
	 */
	public int getCapacity ()
	{
		return capacity;
	}

	/**
	 * @return the didRead
	 */
	public boolean isDidRead ()
	{
		return didRead;
	}

	/**
	 * @return the numberOfCustomers
	 */
	public int getNumberOfCustomers ()
	{
		return numberOfCustomers;
	}
}
