package model;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import model.fillingFunction.AbstractFillingFunction;

public class Solution
{
	private int numberOfVehicles;
	private int vehiclesUsedCount;
	private int numberOfCustomers;
	private Vehicle[] vehicles;
	private int[] customersRoutes;

	/**
	 * Dummy constructor only to be used for initializing TabuList.
	 */
	public Solution ()
	{
	}

	public Solution (Vehicle[] vehicles, int numberOfCustomers)
	{
		if (vehicles == null)
			throw new IllegalArgumentException("The array of vehicles given as a constructor for the solution cannot be null.");

		this.vehicles = vehicles;
		this.numberOfVehicles = vehicles.length;
		vehiclesUsedCount = activeVehicles(null).length;

		this.numberOfCustomers = numberOfCustomers;
		customersRoutes = new int[numberOfCustomers - 1];
		for (int j = 0; j < numberOfCustomers - 1; j++)
			customersRoutes[j] = -1;

		int k = 0;
		for (Vehicle vehicleSpecific : vehicles)
		{
			for (int customer : vehicleSpecific.getCustomersVisiting())
			{
				if (customer == 0)
					continue;
				customersRoutes[customer - 1] = k;
			}
			k++;
		}
	}

	/**
	 * Returns a new reference type solution based on the current solution by executing the move.
	 * 
	 * @param current
	 *          Current solution.
	 * @param move
	 *          Move to be executed.
	 */
	public Solution move (Solution move)
	{
		if (move == null)
			throw new IllegalArgumentException("Both current solution and move must be different from null.");

		if (numberOfCustomers < move.numberOfCustomers)
			throw new IllegalArgumentException("Number of customers in move must be smaller than or equal to the current solution.");

		Vehicle[] currentVehicles = new Vehicle[vehicles.length];
		for (int j = 0; j < vehicles.length; j++)
			currentVehicles[j] = vehicles[j].clone();

		int emptyVehiclesCounter = 0;
		for (Vehicle v : currentVehicles)
		{
			for (Vehicle v2 : move.vehicles)
				for (Integer customer : v2.getCustomersVisiting())
					if (customer == 0)
						continue;
					else if (v.getCustomersVisiting().contains(customer))
						v.RemoveCustomer(customer);

			if (!v.isActive())
			{
				emptyVehiclesCounter++;
				v = null;
			}
		}

		int totalNumberOfVehicles = numberOfVehicles + move.numberOfVehicles - emptyVehiclesCounter;
		Vehicle[] vehicles = new Vehicle[totalNumberOfVehicles];

		int j = 0;
		for (int i = 0; i < numberOfVehicles; i++)
		{
			if (currentVehicles[i].isActive())
			{
				vehicles[j] = currentVehicles[i];
				j++;
			}
		}

		for (int i = 0; i < move.numberOfVehicles; i++)
			vehicles[totalNumberOfVehicles - 1 - i] = move.vehicles[i].clone();

		return new Solution(vehicles, numberOfCustomers);
	}

	/**
	 * Tests if Solution is equal to 'other.' Tests if customers contained in one route are also
	 * contained in the same route in the other solution.
	 */
	@Override
	public boolean equals (Object other)
	{
		if (other == null || this.getClass() != other.getClass())
			return false;

		Solution otherSol = (Solution) other;
		if (this.numberOfVehiclesUsed() != otherSol.numberOfVehiclesUsed())
			return false;

		List<Integer> usedIndexes = new LinkedList<Integer>();
		for (int k = -1; k < this.numberOfVehiclesUsed(); k++)
		{
			int i = 0;
			boolean firstHit = true;
			for (int j = 0; j < numberOfCustomers - 1; j++)
			{
				if (firstHit && customersRoutes[j] == k)
				{
					i = otherSol.customersRoutes[j];
					if (usedIndexes.contains(i))
						return false;
					usedIndexes.add(i);
					firstHit = false;
				}

				if (!firstHit && customersRoutes[j] == k && otherSol.customersRoutes[j] != i)
					return false;
			}
		}
		return true;
	}

	/**
	 * Returns an array of length equal to the number of customers. If entry is -1 the customer is not
	 * visited by any vehicle. Other entries are larger that or equal to 0. If two customers have the
	 * same entry it means they are part of the same route, that is visited by the same vehicle.
	 * 
	 * The depot is not included.
	 * 
	 * @return Array of customer routes.
	 */
	public int[] getCustomerRoutes ()
	{
		return customersRoutes;
	}

	/**
	 * Returns number of vehicles used in current solution.
	 * 
	 * @return Number of vehicles used in current solution.
	 */
	public int numberOfVehiclesUsed ()
	{
		vehiclesUsedCount = activeVehicles(null).length;
		return vehiclesUsedCount;
	}

	/**
	 * Returns a list of the customers a given vehicle visits.
	 * 
	 * @param vehicle
	 *          The vehicle to query.
	 * @return The specific customers.
	 */
	public List<Integer> getCustomersVisited (int vehicle)
	{
		if (vehicle >= vehiclesUsedCount || vehicle < 0)
			throw new IllegalArgumentException("Vehicle index is out of range.");

		return vehicles[vehicle].getCustomersVisiting();
	}

	/**
	 * Returns a list of the customers a list of vehicles visits.
	 * 
	 * @param vehicles
	 *          The list of vehicles to query.
	 * @return The specific customers visited by the queried vehicles.
	 */
	public List<Integer> getCustomersVisited (List<Integer> vehicles)
	{
		List<Integer> customersVisited = new LinkedList<Integer>();
		for (int i : vehicles)
			customersVisited.addAll(getCustomersVisited(i));

		return customersVisited;
	}

	/**
	 * Returns a list of all customers visited by all vehicles.
	 * 
	 * @return The customers visited in this solution.
	 */
	public List<Integer> getCustomersVisited ()
	{
		List<Integer> customersVisited = new LinkedList<Integer>();
		for (int i = 0; i < numberOfCustomers - 1; i++)
			if (customersRoutes[i] > -1)
				customersVisited.add(i + 1);

		return customersVisited;
	}

	/**
	 * Proves solution for validity, that is checks if all customers are visited.
	 * 
	 * @return True if solution is valid, and false otherwise.
	 */
	public boolean isValid ()
	{
		return getCustomersVisited().size() == numberOfCustomers - 1;
	}

	/**
	 * Returns all k-tuple of routes excluding the route containing the targetVehicle.
	 * 
	 * @param targetVehicle
	 *          Vehicle to be excluded.
	 * @param k
	 *          The tuple length.
	 * @return List of lists of customers.
	 */
	public List<List<Integer>> getTupleOfRoutes (Vehicle targetVehicle, int k)
	{
		if (targetVehicle == null)
			throw new IllegalArgumentException("targetVehicle cannot be null.");

		int target = locateVehicle(targetVehicle);

		long binom = Utilities.binom(vehiclesUsedCount - 1, k);

		List<List<Integer>> tupleOfCustomers = new ArrayList<List<Integer>>();
		Vehicle[] activeVehicles = activeVehicles(null);

		byte[] mask = new byte[vehiclesUsedCount];
		Utilities.nextKT(mask, k, target, vehiclesUsedCount);

		for (int j = 0; j < binom; j++)
		{
			List<Integer> tempList = new ArrayList<Integer>();
			for (int i = 0; i < vehiclesUsedCount; i++)
				if (mask[i] == 1)
				{
					List<Integer> l = new LinkedList<Integer>(activeVehicles[i].getCustomersVisiting());
					l.remove(0);// remove depot
					tempList.addAll(l);
				}

			tupleOfCustomers.add(tempList);
			Utilities.nextKT(mask, k, target, vehiclesUsedCount);
		}

		return tupleOfCustomers;
	}

	/**
	 * Returns the position of the specific vehicle in the vehicle array.
	 * 
	 * @param vehicle
	 *          The specific vehicle to look for.
	 * @return The position of the specific vehicle. -1 if it's not part of the solution's vehicle
	 *         array.
	 */
	public int locateVehicle (Vehicle vehicle)
	{
		for (int i = 0; i < numberOfVehicles; i++)
			if (vehicles[i].equals(vehicle))
				return i;

		return -1;
	}

	/**
	 * Returns a new solution that is based on the present solution, but where each customer in the
	 * list is visited by a separate vehicle.
	 * 
	 * @param customersToSeparate
	 *          Customer to visit with separate vehicles.
	 * @return Solution with specified customers visited separately.
	 */
	public Solution visitWithSeparateVehicles (List<Integer> customersToSeparate, AbstractFillingFunction fillingFunction)
	{
		if (customersToSeparate.contains(0))
			throw new IllegalArgumentException("You cannot visit the depot with a separate vehicle.");

		int newLength = customersToSeparate.size() + vehicles.length;
		Vehicle[] existingVehicles = new Vehicle[vehicles.length];
		for (int i = 0; i < vehicles.length; i++)
			existingVehicles[i] = vehicles[i].clone();

		List<Vehicle> newVehicles = new LinkedList<Vehicle>();

		for (int j = 0; j < existingVehicles.length; j++)
		{
			for (Integer customer : customersToSeparate)
			{
				if (existingVehicles[j].ContainsCustomer(customer))
					existingVehicles[j].RemoveCustomer(customer);
			}

			if (existingVehicles[j].getNumbersOfCustomersVisiting() > 1)
				newVehicles.add(existingVehicles[j]);
		}

		int capacity = existingVehicles[0].getCapacity();
		Customers customers = existingVehicles[0].getCustomers();

		List<Integer> customerToAssign = new LinkedList<Integer>(customersToSeparate);

		for (int j = existingVehicles.length; j < newLength; j++)
		{
			newVehicles.add(new Vehicle(capacity, customers, fillingFunction, customerToAssign.get(0)));
			customerToAssign.remove(0);
		}

		return new Solution(newVehicles.toArray(new Vehicle[0]), numberOfCustomers);
	}

	/**
	 * Returns an array of activeVehicles excluding the targetVehicle. Call with null as argument to
	 * obtain list of all vehicles.
	 * 
	 * @param targetVehicle
	 *          The vehicle to be excluded.
	 * @return Array of Vehicles that are active excluding targetVehicle.
	 */
	private Vehicle[] activeVehicles (Vehicle targetVehicle)
	{
		List<Vehicle> activeList = new ArrayList<Vehicle>();

		int counter = 0;
		for (int i = 0; i < numberOfVehicles; i++)
		{
			if (vehicles[i].isActive() && !vehicles[i].equals(targetVehicle))
			{
				activeList.add(vehicles[i]);
				counter++;
			}
		}

		Vehicle[] activeVehicles = new Vehicle[counter];
		activeList.toArray(activeVehicles);

		return activeVehicles;
	}

	public Vehicle[] getVehicles ()
	{
		return vehicles;
	}

	public double getCosts ()
	{
		double cost = 0.0;

		for (Vehicle v : vehicles)
			cost += v.getCost();

		return cost;
	}

	@Override
	public String toString ()
	{
		if (numberOfVehiclesUsed() == 0)
			return "No vehicles in use!";

		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < vehicles.length; i++)
			sb.append("vehicle " + i + ": " + vehicles[i].toString() + "\n");

		return sb.toString();
	}
}
