package model;

import java.util.LinkedList;
import java.util.List;
import java.util.Observable;

import model.fillingFunction.AbstractFillingFunction;
import model.innerSolver.AbstractInnerSolver;

public class TabuSearch extends Observable implements Runnable
{
	private boolean doPrint = false;
	private boolean isFinished = false;
	private boolean isStepping = false;
	private boolean isFullRun = false;

	private int vehicleCapacity;

	// parameters
	private Customers customers;
	private AbstractInnerSolver heuristic;
	private AbstractFillingFunction fillingFunction;
	private int tabuListLength;
	private int iterationLimit;
	private int searchCount;
	private int k_max;
	private int d_max;

	// to be used in algorithm
	private int k;
	private int d;
	private boolean diversify = false;
	private TabuListContainer tabuListContainer;
	private Solution initialSolution;
	private Solution solution;
	private Vehicle targetVehicle;
	private int vehiclesUsed;
	private Solution optimalSolution;
	private long timeLimit = 600000;

	public TabuSearch (DataReader dataReader, Solution initialSolution, AbstractInnerSolver heuristic, AbstractFillingFunction fillingFunction,
			int tabuListLength, int iterationLimit, int k_max, int d_max, boolean doPrint)
	{
		this.customers = new Customers(dataReader);
		this.vehicleCapacity = dataReader.getCapacity();
		this.heuristic = heuristic;
		this.fillingFunction = fillingFunction;
		this.initialSolution = initialSolution;
		solution = initialSolution;
		optimalSolution = initialSolution;
		this.tabuListLength = tabuListLength;
		this.iterationLimit = iterationLimit;
		this.k_max = k_max;
		this.d_max = d_max;
		tabuListContainer = new TabuListContainer(tabuListLength, k_max);
		this.doPrint = doPrint;

		// Preliminary setup
		targetVehicle = Utilities.determineTargetVehicle(solution);
		vehiclesUsed = solution.numberOfVehiclesUsed();
		d = 1;
	}

	/**
	 * Runs the Tabu Search algorithm and returns a solution.
	 * 
	 * @return Solution produced.
	 */
	public void execute ()
	{
		int lowerBound = Math.max(1, (int) Math.ceil(customers.getTotalDemand() / vehicleCapacity));
		vehiclesUsed = solution.numberOfVehiclesUsed();
		if (vehiclesUsed == lowerBound)
			return;

		d = 1;
		targetVehicle = Utilities.determineTargetVehicle(solution);
		int iteration = 1;
		isFinished = true;
		sendNotification();

		long startTime = System.currentTimeMillis();
		
		while (iteration <= iterationLimit && System.currentTimeMillis() - startTime < timeLimit)
		{
			if (doPrint)
				System.out.println("iteration " + iteration + " out of " + iterationLimit);

			diversify = false;
			k = 1;

			while (!diversify && vehiclesUsed > lowerBound && System.currentTimeMillis() - startTime < timeLimit)
			{
				step();
			}

			if (vehiclesUsed == lowerBound)
				return;

			iteration++;
		}

		isFinished = true;
		return;
	}

	public void run ()
	{
		while (true)
		{
			if (isStepping)
			{
				step();
				isStepping = false;
			} else if (isFullRun)
			{
				execute();
				isFullRun = false;
			} else
			{
				sleep();
			}
		}

	}

	private void step ()
	{
		int k_in = k;
		Search();

		searchCount++;
		if(searchCount == 1000)
			diversify = true;

		if (k <= k_in)
			targetVehicle = Utilities.determineTargetVehicle(solution, d);

		if (diversify)
			Diversification();
	}

	private void sleep ()
	{
		try
		{
			Thread.sleep(100);
		} catch (InterruptedException e)
		{
			e.printStackTrace();
		}
	}

	private void Search ()
	{
		List<Integer> customersVisiting = targetVehicle.getCustomersVisiting();
		customersVisiting.remove(0);

		double penalty_star = Double.MAX_VALUE;
		Solution penalty_move = null;
		List<List<Integer>> tupleOfRoutes = solution.getTupleOfRoutes(targetVehicle, k);

		for (int j : customersVisiting)
		{
			for (List<Integer> customersSubList : tupleOfRoutes)
			{
				customersSubList.add(j);
				double penalty = Double.MAX_VALUE;
				Solution move = heuristic.GetSolution(customersSubList);
				int numberVehicles = move.numberOfVehiclesUsed();

				if (numberVehicles < k && numberVehicles > 0)
				{
					solution = solution.move(move);
					updateOptimalSolution();
					sendNotification();

					if (doPrint)
						System.out.println("1: " + solution.numberOfVehiclesUsed() + ", k=" + k);

					k = Math.max(1, k - 1);
					return;
					
				} else if (numberVehicles == k && numberVehicles > 0)
				{
					boolean singleton = targetVehicle.getNumbersOfCustomersVisiting() == 1;

					if (!tabuListContainer.contains(penalty, k) || singleton)
					{
						solution = solution.move(move);
						updateOptimalSolution();
						sendNotification();

						if (doPrint)
							System.out.println("2: " + solution.numberOfVehiclesUsed() + ", k=" + k);

						tabuListContainer.addMove(penalty, k);

						if (singleton)
							k = Math.max(1, k - 1);

						return;
					}
				} else if (numberVehicles == k + 1 && k > 1)
				{
					Vehicle target_bar = Utilities.determineTargetVehicle(move);

					List<Integer> subCustomersVisiting = new LinkedList<Integer>(customersVisiting);
					subCustomersVisiting.remove((Integer) j);
					subCustomersVisiting.addAll(target_bar.getCustomersVisiting());
					subCustomersVisiting.remove((Integer) 0);
					Solution sol = heuristic.GetSolution(subCustomersVisiting);

					penalty = Math.min(sol.getVehicles()[0].fillingFunction(), Utilities.determineTargetVehicle(move, 2).fillingFunction());

					if (sol.numberOfVehiclesUsed() == 1 && !tabuListContainer.contains(penalty, k))
					{
						if (penalty < penalty_star)
						{
							penalty_move = move.move(sol);
							penalty_star = penalty;
						}
					}
				}
			}
		}

		if (penalty_star < Double.MAX_VALUE)
		{
			solution = solution.move(penalty_move);
			sendNotification();
			if (doPrint)
				System.out.println("3: " + solution.numberOfVehiclesUsed() + ", k=" + k + "penalty=" + penalty_star);

			tabuListContainer.addMove(penalty_star, k);
		} else if (k == k_max)
			diversify = true;
		else
			k++;
	}

	private void Diversification ()
	{
		int z = solution.numberOfVehiclesUsed();
		searchCount = 0;

		if (d <= z && d < d_max)
		{
			d++;
			targetVehicle = Utilities.determineTargetVehicle(solution, d);
			if (doPrint)
				System.out.println("Diversify: Target vehicle is now chosen with the " + d + "th lowest value." + " k=" + k);

		} else
		{
			List<Integer> customersToSeparate = new LinkedList<Integer>();
			int l = (int) Math.floor(z / 2);

			List<Integer> tempList = null;
			for (int j = 1; j <= l; j++)
			{
				tempList = Utilities.determineTargetVehicle(solution, j).getCustomersVisiting();
				tempList.remove((Integer) 0);
				customersToSeparate.addAll(tempList);
			}

			tabuListContainer.resetAllTabuList();
			d = 1;
			solution = solution.visitWithSeparateVehicles(customersToSeparate, fillingFunction);
			targetVehicle = Utilities.determineTargetVehicle(solution);

			if (doPrint)
				System.out.println("Diversify: Tabu lists are reset and the following customers are visited by separate vehicles: "
						+ customersToSeparate.toString());
		}
	}

	/**
	 * Method checking whether the current solution uses fewer or same number of vehicles as the
	 * current optimal solution. If the number of vehicles used is the same the costs (i.e. distance
	 * traveled) is compared and the shorter solution is preferred.
	 */
	private void updateOptimalSolution ()
	{
		if (solution.numberOfVehiclesUsed() <= vehiclesUsed
				&& (solution.numberOfVehiclesUsed() < vehiclesUsed || solution.getCosts() < optimalSolution.getCosts()))
		{
			optimalSolution = solution;
			vehiclesUsed = solution.numberOfVehiclesUsed();
			if (doPrint)
				System.out.println("Better solution found! " + optimalSolution.numberOfVehiclesUsed() + ", " + optimalSolution.getCosts());
		}
	}

	@Override
	public String toString ()
	{
		String separator = "\t&";
		return "" + optimalSolution.numberOfVehiclesUsed() + separator + optimalSolution.getCosts() + separator + tabuListLength + separator + iterationLimit + separator
				+ k_max + separator + d_max + separator + fillingFunction.toString();
	}

	public void reset ()
	{
		solution = initialSolution;
		sendNotification();
	}

	public Solution getWorkingSolution ()
	{
		return solution;
	}

	protected void sendNotification ()
	{
		setChanged();
		notifyObservers();
	}

	/**
	 * @return the customers
	 */
	public Customers getCustomers ()
	{
		return customers;
	}

	/**
	 * @return the runFlag
	 */
	public boolean isFinished ()
	{
		return isFinished;
	}

	/**
	 * @return the optimalSolution
	 */
	public Solution getOptimalSolution ()
	{
		return optimalSolution;
	}

	public void setIsStepping (boolean isStepping)
	{
		this.isStepping = isStepping;
	}

	public void setIsFullRun (boolean isFullRun)
	{
		this.isFullRun = isFullRun;
	}
}
