package model;

/**
 * Collection of utilities used for computations in VRPTW. All methods are static.
 * 
 * @author jerome
 * 
 */
public class Utilities
{

	/**
	 * Implementation of binomial coefficient n over k. Returns 0 if k>n or k<0.
	 * 
	 * @param n
	 * @param k
	 * @return n over k
	 */
	public static long binom (int n, int k)
	{
		if (k < 0 || k > n)
			return 0;

		long top = 1;
		long bottom = 1;
		for (int i = 0; i < k; i++)
		{
			top *= (n - i);
			bottom *= (i + 1);
		}
		return (long) top / bottom;
	}

	public static void printv (byte[] mask, int n)
	{
		System.out.printf("{");
		for (int i = 0; i < n; i++)
		{
			if (mask[i] == 1)
				System.out.printf("%d,", 1);// i + 1);
			else
				System.out.printf("%d,", 0);
		}
		System.out.println("}");
	}

	public static boolean next (byte[] mask, int n)
	{
		int i = 0;
		for (; i < n && mask[i] == 1; i++)
			mask[i] = 0;

		if (i < n)
		{
			mask[i] = 1;
			return true;
		}

		return false;
	}

	public static boolean next (byte[] mask, int n, int k)
	{
		// PERFORMANCE HACK: special case for k==1.
		if (k == 1)
		{
			for (int j = 0; j < n; j++)
				if (mask[j] == 1)
				{
					if (j == n - 1)
						return false;

					mask[j] = 0;
					mask[j + 1] = 1;
					return true;
				}
			mask[0] = 1;
			return true;
		}

		int j = 0;
		boolean bool = true;

		while (j != k && bool)
		{
			j = 0;
			bool = next(mask, n);
			if (j >= k)
				continue;
			for (Byte i : mask)
				if (i == 1)
				{
					j++;
					if (j > k)
						break;
				}

		}

		return bool;
	}

	/**
	 * Adopted from Vigo's code. Returns the next mask used for producing subsets.
	 * 
	 * @param kb
	 *          mask.
	 * @param k
	 *          Size of subset.
	 * @param t
	 *          TargetVehicle to be excluded.
	 * @param n
	 *          Size of set.
	 * @return
	 */
	public static boolean nextKT (byte[] kb, int k, int t, int n)
	{
		int i, j, h, index;
		boolean start = true;

		i = index = 0;
		if ((k >= n) || (k == 0))
			return false;

		kb[t] = 0;
		for (i = n - 1; i >= 0; i--)
		{
			if (kb[i] == 1)
			{
				start = false;
				for (j = i + 1; j < n; j++)
					if ((kb[j] == 0) && (j != t))
					{
						kb[j] = 1;
						break;
					}
				if (j != n)
				{
					kb[i] = 0;
					j++;
					for (h = n - 1; h > j;)
					{
						if (kb[h] == 1)
						{
							if ((kb[j] == 0) && (j != t))
							{
								kb[h--] = 0;
								kb[j++] = 1;
							}
							else
								j++;
						}
						else
							h--;
					}
					return true;
				}
			}
		}
		if (start)
		{
			index = i = 0;
			while (i < n)
			{
				if ((i != t) && (index < k))
				{
					kb[i++] = 1;
					index++;
					if (index == k)
						return true;
				}
				else
					kb[i++] = 0;
			}
			return true;
		}
		for (h = 0; h < n; h++)
			kb[h] = 0;
		return false;
	}

	/**
	 * Returns a clone of the vehicle with the lowest value. This is argmin for the phi function of
	 * the article.
	 * 
	 * @param solution
	 *          Solution to be searched for target vehicle.
	 * 
	 * @return The vehicle which route we wish to try to empty.
	 */
	public static Vehicle determineTargetVehicle (Solution solution)
	{
		double min = Double.MAX_VALUE;
		Vehicle target = null;
		for (Vehicle v : solution.getVehicles())
		{
			double d = v.fillingFunction();
			if (d < min)
			{
				target = v;
				min = d;
			}
		}

		return target.clone();
	}

	/**
	 * Returns a clone of the vehicle with the dth-lowest value. This is argmin_dth for the phi
	 * function of the article. If the solution contains less than d vehicles then the vehicle with
	 * the lowest value is simply returned.
	 * 
	 * @param solution
	 *          Solution to be searched for target vehicle.
	 * 
	 * @param d2
	 *          The d-smallest value.
	 * @return The vehicle which route we wish to try to empty.
	 */
	public static Vehicle determineTargetVehicle (Solution solution, int k)
	{
		if (k < 1)
			throw new IllegalArgumentException(
					"You must ask for a vehicle with a positive minimum value. You asked for k=" + k);
		k = k > solution.numberOfVehiclesUsed() ? solution.numberOfVehiclesUsed() : k;

		Vehicle[] heap = new Vehicle[k];
		double k_min = Double.MAX_VALUE;

		for (Vehicle v : solution.getVehicles())
		{
			double d = v.fillingFunction();
			if (d < k_min)
			{
				for (int i = 0; i < k; i++)
				{
					if (heap[i] == null || heap[i].fillingFunction() > d)
					{
						for (int j = k - 1; j > i; j--)
							heap[j] = heap[j - 1];
						heap[i] = v;
						break;
					}
				}
				k_min = heap[k - 1] == null ? Double.MAX_VALUE : heap[k - 1].fillingFunction();
			}
		}

		return heap[k - 1].clone();
	}
}
