package model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import model.fillingFunction.AbstractFillingFunction;

public class Vehicle
{
	private int capacity;
	private int capacityUsed;
	private double cost;
	private boolean costCalculated = false;
	private Customers customers;
	private List<Integer> customersVisiting = new LinkedList<Integer>();
	private int numbersOfCustomersVisiting;
	private List<Double> arrivalTimes;
	private boolean arrivalTimesCalculated = false;
	private AbstractFillingFunction fillingFunction;
	private double fillingFunctionValue;
	private boolean fillingFunctionCalculated = false;

	/**
	 * Class describing a vehicle in the VRPTW.
	 * 
	 * @param capacity
	 *          The capacity of the vehicle.
	 * @param customers
	 *          List of customers the vehicle can serve.
	 * 
	 * @param fillingFunction
	 *          Filling function for a particular vehicle to use.
	 */
	public Vehicle (int capacity, Customers customers, AbstractFillingFunction fillingFunction)
	{
		this.capacity = capacity;
		this.customers = customers;
		this.fillingFunction = fillingFunction;
		customersVisiting.add(0);// Adds depot to list.
		numbersOfCustomersVisiting = 1;// Depot counts as customer.
	}

	/**
	 * Constructs a vehicle that visits a specific customer.
	 * 
	 * @param capacity
	 *          Capacity of vehicle.
	 * @param customers
	 *          Customer database.
	 * @param customerVisiting
	 *          Customer to visit.
	 * 
	 * @param fillingFunction
	 *          Filling function for a particular vehicle to use.
	 */
	public Vehicle (int capacity, Customers customers, AbstractFillingFunction fillingFunction, int customerVisiting)
	{
		this.capacity = capacity;
		this.customers = customers;
		this.fillingFunction = fillingFunction;
		customersVisiting.add(0);// Adds depot to list.
		customersVisiting.add(customerVisiting);
		numbersOfCustomersVisiting = 2;// Depot counts as customer.
	}

	/**
	 * Adds customer to the end of vehicles customer list.
	 * 
	 * @param customer
	 *          The customer to be added.
	 */
	public boolean AssignCustomer (int customer)
	{
		return AssignCustomer(customer, -1);
	}

	/**
	 * Adds a collection of customers to the end of the vehicles customer list.
	 * 
	 * @param customers
	 *          The customers to be added.
	 */
	public boolean AssignCustomer (Collection<Integer> customers)
	{
		boolean bool = true;
		for (Integer j : customers)
			bool &= AssignCustomer(j);

		return bool;
	}

	/**
	 * Adds customer to the end of vehicles customer list at specified index. If index is larger than
	 * the size of the list or negative the customer is added at the end.
	 * 
	 * @param customer
	 *          The customer to be added.
	 * @param index
	 *          The index at which to add the customer.
	 */
	public boolean AssignCustomer (int customer, int index)
	{
		if (index == 0)
			throw new IllegalArgumentException("Index cannot be 0. A vehicle must start in the depot.");
		else if (index > customersVisiting.size())
			throw new IllegalArgumentException("Index cannot be greater than " + customersVisiting.size());
		else if (index < -1)
			throw new IllegalArgumentException("Index cannot be smaller than -1.");

		if (customersVisiting.contains(customer))
			return false;

		int customerDemand = customers.getDemand(customer);
		if (capacityUsed + customerDemand > capacity)
			return false;

		List<Integer> tempCustomersVisiting = new ArrayList<Integer>(customersVisiting);

		if (index == -1)
			tempCustomersVisiting.add(customer);
		else
			tempCustomersVisiting.add(index, customer);

		clearCache();

		List<Double> tempArrivalTimes = calculateArrivalTimes(tempCustomersVisiting);
		if (!checkArrivalTimes(tempCustomersVisiting, tempArrivalTimes))
			return false;

		capacityUsed += customerDemand;
		customersVisiting = tempCustomersVisiting;
		numbersOfCustomersVisiting += 1;
		arrivalTimes = tempArrivalTimes;
		return true;
	}

	public boolean canAssignCustomer (int customer, int index)
	{
		if (customersVisiting.contains(customer) || index == 0 || index < -1)
			return false;

		int customerDemand = customers.getDemand(customer);
		if (capacityUsed + customerDemand > capacity)
			return false;

		List<Integer> tempCustomersVisiting = new ArrayList<Integer>(customersVisiting);

		if (index == -1)
			tempCustomersVisiting.add(customer);
		else
			tempCustomersVisiting.add(index, customer);

		if (!checkArrivalTimes(tempCustomersVisiting, calculateArrivalTimes(tempCustomersVisiting)))
			return false;

		return true;
	}

	/**
	 * Checks a pair of arrivalTimes and customers. arrivalTimes = (cust1, ..., custN, depot).
	 * customersVisiting = (depot, cust1,..., custN).
	 * 
	 * @param arrivalTimes
	 * @return
	 */
	private boolean checkArrivalTimes (List<Integer> customersVisiting, List<Double> arrivalTimes)
	{
		int size = arrivalTimes.size();
		for (int j = 0; j < size - 1; j++)
		{
			int customer = customersVisiting.get(j + 1);
			double arrivalTime = arrivalTimes.get(j);

			if (!(customers.getReadyTime(customer) <= arrivalTime) || !(arrivalTime <= customers.getDueDate(customer)))
				return false;
		}

		if (arrivalTimes.get(size - 1) > customers.getDueDate(0))
			return false;

		return true;
	}

	/**
	 * Return a list of length corresponding to the number of customers that are being visited. First
	 * number is the arrival time at customer one while the last number is the check in time
	 * concluding the tour at the depot.
	 * 
	 * If vehicle never leaves the depot (depot), that is (0), is returned.
	 * 
	 * @param customersVisiting
	 * @return arrivalTimes Arrival Times (cust1, cust2,..., custN, depot).
	 */
	private List<Double> calculateArrivalTimes (List<Integer> customersVisiting)
	{
		if (customersVisiting.isEmpty())
			throw new IllegalArgumentException("List of customers for which to calculate arrival times cannot be empty.");

		int size = customersVisiting.size();
		List<Double> arrivalTimes = new LinkedList<Double>();
		if (size == 1)
		{
			arrivalTimes.add(0.0);
			return arrivalTimes;
		}

		double time = 0.0;

		int startCustomer = customersVisiting.get(0);
		int customer1 = startCustomer;
		int customer2 = customersVisiting.get(1);

		for (int j = 0; j < size - 1; j++)
		{
			customer2 = customersVisiting.get(j + 1);

			time += customers.getServiceTime(customer1) + customers.Distance(customer1, customer2);
			time = Math.max(time, customers.getReadyTime(customer2));
			arrivalTimes.add(time);

			customer1 = customer2;
		}
		// Adds the time back to depot.
		time += customers.getServiceTime(customer2) + customers.Distance(customer2, startCustomer);
		arrivalTimes.add(time);

		return arrivalTimes;
	}

	private List<Double> calculateArrivalTimes ()
	{
		if (!arrivalTimesCalculated)
		{
			this.arrivalTimes = calculateArrivalTimes(customersVisiting);
			arrivalTimesCalculated = true;
		}

		return arrivalTimes;

	}

	/**
	 * Returns the time for which the vehicle is back at the depot.
	 * 
	 * @return Arrival time at depot.
	 */
	public double checkInTime ()
	{
		return arrivalTimes.get(arrivalTimes.size() - 1);
	}

	/**
	 * Removes a collection of customers from the given vehicle. Only customers that are being visited
	 * can be removed.
	 * 
	 * @param customers
	 *          Collection of customers to remove.
	 * @return 'True' if customers are successfully removed.
	 */
	public boolean RemoveCustomer (Collection<Integer> customers)
	{
		boolean bool = true;
		for (Integer j : customers)
			bool &= RemoveCustomer(j);

		return bool;
	}

	/**
	 * Removes a given customer from the vehicles route. Only customers that are actually being
	 * visited can be removed.
	 * 
	 * @param customer
	 *          Customer to remove.
	 * @return Returns 'true' if customer is successfully removed.
	 */
	public boolean RemoveCustomer (int customer)
	{
		if (customer == 0)
			throw new IllegalArgumentException("Customer cannot be 0. A vehicle must start in the depot.");

		// How expensive is this check?
		if (!customersVisiting.contains(customer))
			throw new IllegalArgumentException("Cannot remove a customer that is not already being visited.");

		customersVisiting.remove((Integer) customer);
		clearCache();

		capacityUsed -= customers.getDemand(customer);
		numbersOfCustomersVisiting -= 1;
		calculateArrivalTimes();

		return true;
	}

	/**
	 * Clears the cache.
	 */
	private void clearCache ()
	{
		arrivalTimesCalculated = false;
		costCalculated = false;
		fillingFunctionCalculated = false;
	}

	/**
	 * Returns the number of customers being visited including the depot.
	 * @return the numbersOfCustomersVisiting
	 */
	public int getNumbersOfCustomersVisiting ()
	{
		return numbersOfCustomersVisiting;
	}

	public void ClearListOfCustomersVisiting ()
	{
		customersVisiting.clear();
		customersVisiting.add(0);// Adds depot to list.

		capacityUsed = 0;
		clearCache();
	}

	/**
	 * Returns the cost of operating this vehicle with the present list of customers. Uses caching.
	 * 
	 * @return The cost.
	 */
	public double getCost ()
	{
		if (!costCalculated)
			CalculateCost();

		return cost;
	}

	private void CalculateCost ()
	{
		cost = CalculateCost(customersVisiting);
		costCalculated = true;
	}

	/**
	 * Calculates the cost for traveling between the customers in the given order in the given in the
	 * list.
	 * 
	 * @param customerList
	 *          The list of customers.
	 * @return The cost of traveling between the customers in the given order.
	 */
	private double CalculateCost (List<Integer> customerList)
	{
		double costSpecific = 0.0;
		int cust1 = customerList.get(0);
		for (int cust2 : customerList)
		{
			costSpecific += customers.Distance(cust1, cust2);
			cust1 = cust2;
		}

		return costSpecific + customers.Distance(customerList.get(0), cust1);
	}

	/**
	 * @return the costCalculated
	 */
	public boolean isCostCalculated ()
	{
		return costCalculated;
	}

	public boolean ContainsCustomer (int customer)
	{
		return customersVisiting.contains(customer);
	}

	/**
	 * @return the capacity
	 */
	public int getCapacity ()
	{
		return capacity;
	}

	/**
	 * Returns the customer class associated with the vehicle.
	 * 
	 * @return Customer class associated with the vehicle.
	 */
	public Customers getCustomers ()
	{
		return customers;
	}

	/**
	 * Returns a new list of the customer, represented by Integers, visited by this particular
	 * vehicle. The depot is included.
	 * 
	 * @return the customersVisiting
	 */
	public List<Integer> getCustomersVisiting ()
	{
		return new LinkedList<Integer>(customersVisiting);
	}

	/**
	 * @param customersVisiting
	 *          the customersVisiting to set
	 */
	public void setCustomersVisiting (List<Integer> customersVisiting)
	{
		this.customersVisiting = customersVisiting;
	}

	/**
	 * @return the capacityUsed
	 */
	public int getCapacityUsed ()
	{
		return capacityUsed;
	}

	/**
	 * @return the capacityUsed
	 */
	public int getSpareCapacity ()
	{
		return capacity - capacityUsed;
	}

	/**
	 * @return the arrivalTimes
	 */
	public List<Double> getArrivalTimes ()
	{
		if (!arrivalTimesCalculated)
			calculateArrivalTimes();

		return new LinkedList<Double>(arrivalTimes);
	}

	public double getArrivalTimes (int customer)
	{
		if (customer == 0)
			return arrivalTimes.get(numbersOfCustomersVisiting - 1);// return arrival time for depot.

		if (customer < 0 || customer > numbersOfCustomersVisiting)
			throw new IllegalArgumentException("Customer number must be between 0 and the number of customers that are visited.");

		return arrivalTimes.get(customer - 1);
	}

	/**
	 * Returns true if vehicle leaves depot and false otherwise.
	 * 
	 * @return true if vehicle leaves depot and false otherwise.
	 */
	public boolean isActive ()
	{
		return numbersOfCustomersVisiting != 1;
	}

	/**
	 * Filling function described in Potvin-paper.
	 * 
	 * @return The value of the filling function.
	 */
	public double fillingFunction ()
	{
		if (!fillingFunctionCalculated)
		{
			calculateArrivalTimes();

			fillingFunctionValue = fillingFunction.calculate(customersVisiting, numbersOfCustomersVisiting, arrivalTimes, capacityUsed);
		}

		return fillingFunctionValue;
	}

	/**
	 * Calculates the increase in cost from adding the given customer at the given index.
	 * 
	 * @param customer
	 *          The customer to be added.
	 * @param index
	 *          The index at which to add the customer.
	 * @return The increase in cost.
	 */
	public double getDetourCost (int customer, int index)
	{
		if (!canAssignCustomer(customer, index))
			throw new IllegalArgumentException("Customer has to be assignable to vehicle at the given index.");

		List<Integer> tempCustomerList = new LinkedList<Integer>(customersVisiting);
		tempCustomerList.add(customer);

		return CalculateCost(tempCustomerList) - getCost();
	}

	/**
	 * Calculates the delay of checking in at the depot implied by adding the given customer at the
	 * given index.
	 * 
	 * @param customer
	 *          The customer to be added.
	 * @param index
	 *          The index at which to add the customer.
	 * @return The delay of check in time.
	 */
	public double getDelay (int customer, int index)
	{
		if (!canAssignCustomer(customer, index))
			throw new IllegalArgumentException("Customer has to be assignable to vehicle at the given index.");

		List<Integer> tempCustomerList = new LinkedList<Integer>(customersVisiting);
		tempCustomerList.add(customer);

		return calculateArrivalTimes(tempCustomerList).get(numbersOfCustomersVisiting) - getArrivalTimes().get(numbersOfCustomersVisiting - 1);
	}

	@Override
	public String toString ()
	{
		List<Integer> tempCustomersVisiting = getCustomersVisiting();
		tempCustomersVisiting.remove(0);
		String s = "{";
		for (Integer customer : tempCustomersVisiting)
			s += customer.toString() + ", ";

		return tempCustomersVisiting.isEmpty() ? s + "}" : s.substring(0, s.length() - 2) + "}";
	}

	@Override
	public Vehicle clone ()
	{
		Vehicle clone = new Vehicle(capacity, customers, fillingFunction);
		clone.AssignCustomer(customersVisiting);

		return clone;
	}

	@Override
	public boolean equals (Object other)
	{
		if (other == null || other.getClass() != Vehicle.class)
			return false;

		return equals((Vehicle) other);
	}

	public boolean equals (Vehicle v)
	{
		if (v == null || numbersOfCustomersVisiting != v.getNumbersOfCustomersVisiting())
			return false;

		for (Integer customer : customersVisiting)
		{
			if (!v.ContainsCustomer(customer))
				return false;
		}

		return true;
	}
}
