package model.innerSolver;

import java.util.LinkedList;
import java.util.List;

import model.Customers;
import model.Solution;
import model.Vehicle;
import model.fillingFunction.AbstractFillingFunction;

/**
 * This class presents a heuristic for finding solution using the Solomon method described in Potvin
 * paper p. 332, section 2. It uses the customer farthest away from the depot as a seed and
 * constructs the route accordingly.
 * 
 * It can be used stand-alone or to initialized the parallel approved described by Potvin.
 * 
 * @author jerome
 * 
 */
public class SolomonHeuristic extends AbstractInnerSolver
{
	private double alpha_1 = 0.5;
	private double alpha_2 = 1.0 - alpha_1;
	
	/**
	 * Default constructor setting both alpha_1 and alpha_2 to 0.5.
	 * 
	 * @param customers
	 * @param capacity
	 * @param fillingFunction Filling function to use for the vehicles that are being constructed.
	 */
	public SolomonHeuristic (Customers customers, int capacity, AbstractFillingFunction fillingFunction)
	{
		super(customers, capacity, fillingFunction);
	}

	/**
	 * Constructs an Solomon heuristic with alpha_1 and alpha_2 specified by the user.
	 * 
	 * @param customers
	 * @param capacity
	 * @param alpha_1
	 * @param alpha_2
	 */
	public SolomonHeuristic (Customers customers, int capacity, AbstractFillingFunction fillingFunction, double alpha_1, double alpha_2)
	{
		super(customers, capacity, fillingFunction);

		if (alpha_1 + alpha_2 != 1 || alpha_1 < 0 || alpha_2 < 0)
			throw new IllegalArgumentException(
					"Both alpha_1 and alpha_2 need to be positive and their sum has to be 1. They were; "
							+ alpha_1 + "and " + alpha_2);

		this.alpha_1 = alpha_1;
		this.alpha_2 = alpha_2;
	}

	/**
	 * Returns an initial solution where all the customers are visited. The solution is produced using
	 * Solomon's heuristic.
	 */
	@Override
	public Solution GetInitialSolution ()
	{
		return GetSolution(customers.getListOfCustomers());
	}

	/**
	 * Finds the optimal customer with respect to the criteria described in Potvin-paper.
	 * 
	 * @param workList
	 *          Worklist to process, that is unassigned customers.
	 * @param vehicle
	 *          Vehicle to add customers to.
	 * @return Returns the customer to be added if there is a feasible customer. Returns -1 otherwise.
	 */
	private int[] findOptimalCustomer (List<Integer> workList, Vehicle vehicle)
	{
		int workSize = workList.size();
		Object[][] opt = new Object[workSize][3];

		int counter = 0;
		double d_optimal = Double.MAX_VALUE;

		// Finds optimal index for a given customer and remembers the weighted average.
		for (int customer : workList)
		{
			d_optimal = Double.MAX_VALUE;
			int index_optimal = 0;
			for (int index = 1; index <= vehicle.getNumbersOfCustomersVisiting(); index++)
			{
				if (!vehicle.canAssignCustomer(customer, index))
					continue;

				double d = alpha_1 * vehicle.getDelay(customer, index) + alpha_2
						* vehicle.getDetourCost(customer, index);
				if (d < d_optimal)
				{
					d_optimal = d;
					index_optimal = index;
				}
			}
			if (d_optimal == Double.MAX_VALUE)
				continue;

			opt[counter][0] = customer;
			opt[counter][1] = index_optimal;
			opt[counter][2] = d_optimal;
			counter++;
		}

		return secondCriterion(opt, workSize);
	}

	/**
	 * Evaluates a second criteria from which to choose the specific customer to implement. In the
	 * particular case it is simply the minimum of an weighted between delay and detour.
	 * 
	 * @param optimals
	 *          Double array consisting of objects. First coordinate lists the customer, while second
	 *          coordinate is the index optimal weighted average.
	 * @param workSize
	 *          The number of customers.
	 * @return Returns an array with customer, optimal index and optimal value.
	 */
	private int[] secondCriterion (Object[][] optimals, int workSize)
	{
		int[] optimalContainer = new int[2];

		double optimalDistance = Double.MAX_VALUE;
		int optimalCustomer = 0;

		for (int i = 0; i < workSize; i++)
		{
			Double customerSpecificOptimum = optimals[i][2] != null ? (Double) optimals[i][2]
					: Double.MAX_VALUE;

			if (customerSpecificOptimum < optimalDistance)
			{
				optimalDistance = customerSpecificOptimum;
				optimalCustomer = i;
			}
		}

		if (optimalDistance == Double.MAX_VALUE)
			return null;

		for (int k = 0; k < 2; k++)
			optimalContainer[k] = (Integer) optimals[optimalCustomer][k];

		return optimalContainer;
	}

	/**
	 * Finds and places the optimal customer according to the criteria described in Potvin-paper.
	 * 
	 * @param workList
	 *          Worklist to process, that is unassigned customers.
	 * @param vehicle
	 *          Vehicle to add customers to.
	 * 
	 * @return Returns true if it is still feasible to assign further customer to the particular
	 *         vehicle and false otherwise.
	 */
	private boolean placeOptimalCustomer (List<Integer> workList, Vehicle vehicle)
	{
		int[] optimalContainer = findOptimalCustomer(workList, vehicle);

		if (optimalContainer == null)
			return false;

		if ((Integer) optimalContainer[0] != -1)
			placeCustomer(vehicle, optimalContainer[0], optimalContainer[1], workList);

		return true;
	}

	/**
	 * Adds customer to vehicle and remove him/her from the worklist.
	 * 
	 * @param vehicle
	 *          Vehicle which to add customer to.
	 * @param seedCustomer
	 *          Customer to be added/removed.
	 * @param workList
	 *          Worklist to remove customer from.
	 */
	private void placeCustomer (Vehicle vehicle, int customer, int index, List<Integer> workList)
	{
		vehicle.AssignCustomer(customer, index);
		workList.remove((Object) customer);
	}

	@Override
	public Solution GetSolution (List<Integer> workList)
	{
		workList = new LinkedList<Integer>(workList);
		
		if (workList.get(0) == 0)
			workList.remove(0); // remove depot.

		if (workList.contains((Integer) 0))
			throw new IllegalArgumentException("Depot is contained multiple times in worklist");

		double distance_max = 0.0;
		int seedCustomer = 0;
		for (int customer : workList)
		{
			if (customers.Distance(customer) > distance_max)
			{
				seedCustomer = customer;
				distance_max = customers.Distance(customer);
			}
		}

		Vehicle vehicle = new Vehicle(capacity, customers, fillingFunction);
		placeCustomer(vehicle, seedCustomer, 1, workList);

		List<Vehicle> vehicles = new LinkedList<Vehicle>();
		
		if(workList.isEmpty())
			vehicles.add(vehicle);

		while (!workList.isEmpty())
		{
			while (placeOptimalCustomer(workList, vehicle))
			{
			}

			vehicles.add(vehicle);
			vehicle = new Vehicle(capacity, customers, fillingFunction);
		}

		return new Solution(vehicles.toArray(new Vehicle[0]), customers.getNumberOfCustomers());
	}
}
