package test;

import java.util.LinkedList;
import java.util.List;

import model.Solution;
import model.Utilities;
import model.Vehicle;
import model.fillingFunction.InitialFillingFunction;
import model.innerSolver.SolomonHeuristic;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class Solution_Test extends AbstractTestCase
{
	private SolomonHeuristic solomonHeuristic;
	private InitialFillingFunction fillingFunction;
	
	@Before
	public void setUp () throws Exception
	{
		baseSetUp();
		solomonHeuristic = new SolomonHeuristic(customersFull, 200, new InitialFillingFunction(customersFull, 0.5));
		fillingFunction = new InitialFillingFunction(customersMedium, 0.5);
	}

	@Test
	public void testCustomerRoutes ()
	{
		Assert.assertEquals(0, solution.getCustomerRoutes()[1]);
		Assert.assertEquals(0, solution.getCustomerRoutes()[2]);
		Assert.assertEquals(0, solution.getCustomerRoutes()[12]);
		Assert.assertEquals(1, solution.getCustomerRoutes()[3]);
		Assert.assertEquals(-1, solution.getCustomerRoutes()[4]);
		Assert.assertEquals(-1, solution.getCustomerRoutes()[5]);
		Assert.assertEquals(-1, solution.getCustomerRoutes()[15]);
	}

	@Test
	public void testEquals ()
	{
		Assert.assertEquals(true, solution.equals(solution));
		Assert.assertEquals(false, solution.equals(null));

		Vehicle[] vehicles2 = new Vehicle[2];
		Vehicle vehicle3 = new Vehicle(200, customersFull, fillingFunction);
		Vehicle vehicle4 = new Vehicle(200, customersFull, fillingFunction);
		vehicle4.AssignCustomer(2);
		vehicle4.AssignCustomer(3);
		vehicle4.AssignCustomer(13);
		vehicle3.AssignCustomer(4);
		vehicles2[0] = vehicle3;
		vehicles2[1] = vehicle4;

		Solution solution2 = new Solution(vehicles2, customersFull.getNumberOfCustomers());
		Assert.assertEquals(true, solution.equals(solution2));

		Solution solution3 = new Solution();
		Assert.assertEquals(false, solution.equals(solution3));
		Assert.assertEquals(true, solution3.equals(solution3));
	}

	@Test
	public void testLocateVehicle ()
	{
		Solution sol = solomonHeuristic.GetInitialSolution();
		for (int i = 0; i < sol.numberOfVehiclesUsed(); i++)
		{
			Vehicle v = sol.getVehicles()[i];
			int position = sol.locateVehicle(v);
			Assert.assertEquals(i, position);
		}
	}

	@Test
	public void testCustomersVisited ()
	{
		Assert.assertEquals(4, solution.getCustomersVisited().size());
		Assert.assertEquals(true, solution.getCustomersVisited().contains((int) 2));
		Assert.assertEquals(true, solution.getCustomersVisited().contains((int) 3));
		Assert.assertEquals(true, solution.getCustomersVisited().contains((int) 4));
		Assert.assertEquals(false, solution.getCustomersVisited().contains((int) 5));
	}

	@Test
	public void testCustomersVisited2 ()
	{
		Assert.assertEquals(100, solomonHeuristic.GetInitialSolution().getCustomersVisited().size());
	}

	@Test
	public void testGetTupleOfRoutes ()
	{
		Assert.assertEquals(1, solution.getTupleOfRoutes(solution.getVehicles()[0], 1).size());
		Assert.assertEquals(1, solution.getTupleOfRoutes(solution.getVehicles()[1], 1).size());
		Solution sol = solomonHeuristic.GetInitialSolution();
		Vehicle[] vehicles = sol.getVehicles();
		int vehiclesUsed = sol.numberOfVehiclesUsed();

		for (int k = 1; k <= 5; k++)
			Assert.assertEquals(Utilities.binom(vehiclesUsed - 1, k), sol
					.getTupleOfRoutes(vehicles[0], k).size());
	}

	@Test
	public void testGetVehicles ()
	{
		Assert.assertEquals(2, solution.getVehicles().length);
	}

	@Test
	public void testPrint ()
	{
		Assert.assertEquals(37, solution.toString().length());
	}

	@Test
	public void testGetCustomerRoutes ()
	{
		Assert.assertEquals(100, solution.getCustomerRoutes().length);
	}

	@Test
	public void testIsValid ()
	{
		Assert.assertFalse(solution.isValid());
		Assert.assertTrue(solomonHeuristic.GetInitialSolution().isValid());
	}

	@Test
	public void testMove ()
	{
		Solution solution2 = solomonHeuristic.GetInitialSolution();
		Solution afterMove = solution2.move(solution);
		Assert.assertEquals(solution2.numberOfVehiclesUsed() + 2, afterMove.numberOfVehiclesUsed());
		Assert.assertTrue(afterMove.getVehicles()[21].ContainsCustomer(4)
				&& afterMove.getVehicles()[22].ContainsCustomer(2)
				&& afterMove.getVehicles()[22].ContainsCustomer(3)
				&& afterMove.getVehicles()[22].ContainsCustomer(13));
		Assert.assertTrue(afterMove.isValid());
	}

	@Test
	public void testMove2 ()
	{
		Solution sol = solomonHeuristic.GetInitialSolution();
		Vehicle v = new Vehicle(200, customersFull, fillingFunction);
		v.AssignCustomer(65);
		v.AssignCustomer(90);
		v.AssignCustomer(1);
		Solution move = new Solution(new Vehicle[] { v }, 101);
		Solution afterMove = sol.move(move);
		Assert.assertEquals(22, afterMove.numberOfVehiclesUsed());
		Assert.assertTrue(afterMove.isValid());
		Assert.assertEquals(4, afterMove.getVehicles()[21].getCustomersVisiting().size());
	}
	
	@Test
	public void testSolutionImmutable ()
	{
		Solution sol = solomonHeuristic.GetInitialSolution();
		Vehicle v = new Vehicle(200, customersFull, fillingFunction);
		v.AssignCustomer(65);
		v.AssignCustomer(90);
		v.AssignCustomer(1);
		Solution move = new Solution(new Vehicle[] { v }, 101);
		Solution afterMove = sol.move(move);
		Assert.assertFalse(afterMove == sol);
		Assert.assertTrue(solomonHeuristic.GetInitialSolution().equals(sol));
		List<Integer> customerToSeperate = new LinkedList<Integer>();
		customerToSeperate.add(1);
		sol.visitWithSeparateVehicles(customerToSeperate, fillingFunction);
		Assert.assertTrue(solomonHeuristic.GetInitialSolution().equals(sol));
	}

	@Test
	public void testVisitWithSeparateVehicles ()
	{
		Solution sol = solomonHeuristic.GetInitialSolution();
		List<Integer> customerToSeparate = new LinkedList<Integer>();
		customerToSeparate.add(1);
		customerToSeparate.add(2);
		customerToSeparate.add(3);
		Solution sol2 = sol.visitWithSeparateVehicles(customerToSeparate, fillingFunction);
		Assert.assertEquals(sol.numberOfVehiclesUsed() + 3, sol2.numberOfVehiclesUsed());
	}
	
	@Test
	public void testVisitWithSeparateVehiclesClone ()
	{
		Solution solution1 = solomonHeuristic.GetInitialSolution();
		int numberOfVehiclesUsed = solution1.numberOfVehiclesUsed();
		Solution solution2 = solution1;
		List<Integer> customersToSeparate = new LinkedList<Integer>();
		customersToSeparate.add(16);
		customersToSeparate.add(86);
		customersToSeparate.add(71);
		solution1 = solution1.visitWithSeparateVehicles(customersToSeparate, fillingFunction);
		Assert.assertEquals(numberOfVehiclesUsed, solution2.numberOfVehiclesUsed());
    Assert.assertTrue(solution1.getVehicles()[19].ContainsCustomer(16));
    Assert.assertTrue(solution1.getVehicles()[20].ContainsCustomer(86));
    Assert.assertTrue(solution1.getVehicles()[21].ContainsCustomer(71));
	}
}
