package view;

import java.awt.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Observable;

import javax.swing.*;

import control.InteractiveController;

import model.*;

/**
 * A Swing component that displays a 2D array of pixels.
 * 
 * The pixels can be updated one at a time.
 * 
 * The pixels can be displayed with a magnification factor that can be changed.
 * 
 * To avoid confusion all sizes and coordinates are given in the source coordinate system.
 */

public class AreaPanel extends JPanel
{
	private Solution solution;
	private Customers customers;

	private List<Color> colorList;
	private int colorListSize;

	/**
	 * Used for serialization purposes
	 */
	public static final long serialVersionUID = 100000;

	/**
	 * Construct a new AreaPanel to show simulation in
	 * 
	 * @param solution
	 *          the simulation we observe.
	 */
	public AreaPanel (TabuSearch ts)
	{
		assert (ts != null);
		customers = ts.getCustomers();

		colorList = new ArrayList<Color>();
		colorList.add(Color.CYAN);
		colorList.add(Color.GREEN);
		colorList.add(Color.MAGENTA);
		colorList.add(Color.ORANGE);
		colorList.add(Color.PINK);
		colorList.add(Color.RED);
		colorList.add(Color.YELLOW);
		colorList.add(Color.BLUE);
		colorList.add(Color.BLACK);
		colorListSize = colorList.size();

		update(ts);
	}

	/**
	 * Make sure that changes are drawn.
	 * 
	 * @param simulation
	 *          The update method invoked by the WorldView
	 */
	public void update (Observable observable)
	{
		if (observable.getClass() == TabuSearch.class)
			solution = ((TabuSearch) observable).getWorkingSolution();
		else if (observable.getClass() == InteractiveController.class)
			solution = ((InteractiveController) observable).getTs().getWorkingSolution();

		repaint();
	}

	/**
	 * Command: Paint the depot and all the vehicles
	 * 
	 * @param g
	 *          the graphics component to use when painting
	 */
	public void paintComponent (Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g2d = (Graphics2D) g;
		g2d.setStroke(new BasicStroke(2.0f));
		g2d.setColor(Color.BLACK);

		int start = 10;
		int end = 610;
		int scale = (end - start) / 100;
		g2d.drawLine(start, start, end, start);
		g2d.drawLine(end, start, end, end);
		g2d.drawLine(end, end, start, end);
		g2d.drawLine(start, end, start, start);

		drawCustomers(g2d, start, scale);
		drawRoutes(g2d, start, end);
	}

	private void drawCustomers (Graphics2D g2d, int start, int scale)
	{
		g2d.setStroke(new BasicStroke(2.0f));
		// Print depot
		g2d.setColor(Color.RED);
		int depotSize = 10;
		g2d.fillOval(start + scale * customers.getCustomerCoordinates(0)[0] - depotSize / 2, start + scale * customers.getCustomerCoordinates(0)[1]
				- depotSize / 2, depotSize, depotSize);

		// Print all customers
		g2d.setColor(Color.BLACK);
		for (int i = 1; i < customers.getNumberOfCustomers(); i++)
		{
			int[] coordinates = customers.getCustomerCoordinates(i);
			g2d.drawRect(start + scale * coordinates[0] - 2, start + scale * coordinates[1] - 2, 4, 4);
			g2d.drawString(((Integer) i).toString(), start + scale * coordinates[0] - 2, start + scale * coordinates[1] - 2);
		}
	}

	private void drawRoutes (Graphics2D g2d, int start, int end)
	{
		g2d.setStroke(new BasicStroke(1.0f));
		int colorCount = 0;

		for (Vehicle v : solution.getVehicles())
		{
			int size = v.getNumbersOfCustomersVisiting();
			int customer1 = v.getCustomersVisiting().get(0);
			Color colorSpecific = colorList.get(colorCount % colorListSize);
			colorCount++;

			for (int i = 1; i < size; i++)
			{
				int customer2 = v.getCustomersVisiting().get(i);

				drawLine(g2d, start, end, customers.getCustomerCoordinates(customer1)[0], customers.getCustomerCoordinates(customer1)[1], customers
						.getCustomerCoordinates(customer2)[0], customers.getCustomerCoordinates(customer2)[1], colorSpecific);

				customer1 = customer2;
			}
			// Add line back to depot
			drawLine(g2d, start, end, customers.getCustomerCoordinates(0)[0], customers.getCustomerCoordinates(0)[1], customers.getCustomerCoordinates(v
					.getCustomersVisiting().get(size - 1))[0], customers.getCustomerCoordinates(v.getCustomersVisiting().get(size - 1))[1], colorSpecific);
		}
	}

	private void drawLine (Graphics2D g2d, int start, int end, int x1, int y1, int x2, int y2, Color c)
	{
		g2d.setColor(c);
		int scale = (end - start) / 100;
		g2d.drawLine(start + scale * x1, start + scale * y1, start + scale * x2, start + scale * y2);
	}
}
